//=================================================================================================
/*
    Copyright (C) 2015-2017 Sky kit authors united with omega. <http://omega.gg/about>

    Author: Benjamin Arnaud. <http://bunjee.me> <bunjee@omega.gg>

    This file is part of the SkCore module of Sky kit.

    - GNU General Public License Usage:
    This file may be used under the terms of the GNU General Public License version 3 as published
    by the Free Software Foundation and appearing in the LICENSE.md file included in the packaging
    of this file. Please review the following information to ensure the GNU General Public License
    requirements will be met: https://www.gnu.org/licenses/gpl.html.
*/
//=================================================================================================

#include "WControllerFile.h"

#ifndef SK_NO_CONTROLLERFILE

// Qt includes
#include <QCoreApplication>
#include <QDir>
#include <qtlockedfile>

// Sk includes
#include <WControllerApplication>
#include <WControllerNetwork>
#include <WLocalObject>
#include <WFileWatcher>
#include <WCache>
#include <WThreadActions>

// Private includes
#include <private/WFileWatcher_p>

// Namespaces
using namespace QtLP_Private;

W_INIT_CONTROLLER(WControllerFile)

//-------------------------------------------------------------------------------------------------
// Static variables

static const int CONTROLLERFILE_TIMEOUT = 10000; // 10 seconds

//=================================================================================================
// WControllerFileAction
//=================================================================================================

/* virtual */ WAbstractThreadReply * WControllerFileAction::createReply() const
{
    return new WControllerFileReply;
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFileAction::controllerReply()
{
    return qobject_cast<WControllerFileReply *> (reply());
}

//=================================================================================================
// WControllerFileWrite
//=================================================================================================

class WControllerFileWrite : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList fileNames;

    QList<QByteArray> datas;
};

/* virtual */ bool WControllerFileWrite::run()
{
    for (int i = 0; i < fileNames.count(); i++)
    {
        WControllerFile::writeFile(fileNames.at(i), datas.at(i));
    }

    return true;
}

//=================================================================================================
// WControllerFileRename
//=================================================================================================

class WControllerFileRename : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList oldPaths;
    QStringList newPaths;
};

/* virtual */ bool WControllerFileRename::run()
{
    for (int i = 0; i < oldPaths.count(); i++)
    {
        WControllerFile::renameFile(oldPaths.at(i), newPaths.at(i));
    }

    return true;
}

//=================================================================================================
// WControllerFileCopy
//=================================================================================================

class WControllerFileCopy : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList fileNames;
    QStringList newNames;
};

/* virtual */ bool WControllerFileCopy::run()
{
    for (int i = 0; i < fileNames.count(); i++)
    {
        WControllerFile::copyFile(fileNames.at(i), newNames.at(i));
    }

    return true;
}

//=================================================================================================
// WControllerFileDelete
//=================================================================================================

class WControllerFileDelete : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList paths;
};

/* virtual */ bool WControllerFileDelete::run()
{
    foreach (const QString & path, paths)
    {
        WControllerFile::deleteFile(path);
    }

    return true;
}

//=================================================================================================
// WControllerFileCreateFolders
//=================================================================================================

class WControllerFileCreateFolders : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList paths;
};

/* virtual */ bool WControllerFileCreateFolders::run()
{
    foreach (const QString & path, paths)
    {
        if (WControllerFile::createFolder(path) == false)
        {
            qWarning("WControllerFileCreateFolders::run: Failed to create folder %s.", path.C_STR);
        }
    }

    return true;
}

//=================================================================================================
// WControllerFileDeleteFolders
//=================================================================================================

class WControllerFileDeleteFolders : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList paths;

    bool recursive;
};

/* virtual */ bool WControllerFileDeleteFolders::run()
{
    foreach (const QString & path, paths)
    {
        if (WControllerFile::deleteFolder(path, recursive) == false)
        {
            qWarning("WControllerFileDeleteFolders::run: Failed to delete folder %s.", path.C_STR);
        }
    }

    return true;
}

//=================================================================================================
// WControllerFileDeleteFoldersContent
//=================================================================================================

class WControllerFileDeleteFoldersContent : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList paths;

    bool recursive;
};

/* virtual */ bool WControllerFileDeleteFoldersContent::run()
{
    foreach (const QString & path, paths)
    {
        if (WControllerFile::deleteFolderContent(path, recursive) == false)
        {
            qWarning("WControllerFileDeleteFoldersContent::run: Failed to delete folder content %s"
                     ".", path.C_STR);
        }
    }

    return true;
}

//=================================================================================================
// WControllerFileCreatePaths
//=================================================================================================

class WControllerFileCreatePaths : public WControllerFileAction
{
    Q_OBJECT

protected: // WAbstractThreadAction implementation
    /* virtual */ bool run();

public: // Variables
    QStringList paths;
};

/* virtual */ bool WControllerFileCreatePaths::run()
{
    QDir dir;

    foreach (const QString & path, paths)
    {
        if (dir.mkpath(path) == false)
        {
            qWarning("WControllerFileCreatePaths::run: Failed to create path %s.", path.C_STR);
        }
    }

    return true;
}

//=================================================================================================
// WControllerFileReply
//=================================================================================================

/* virtual */ void WControllerFileReply::onCompleted(bool ok)
{
    emit actionComplete(ok);
}

//=================================================================================================
// WControllerFilePrivate
//=================================================================================================

WControllerFilePrivate::WControllerFilePrivate(WControllerFile * p) : WControllerPrivate(p) {}

/* virtual */ WControllerFilePrivate::~WControllerFilePrivate()
{
    W_CLEAR_CONTROLLER(WControllerFile);
}

//-------------------------------------------------------------------------------------------------

void WControllerFilePrivate::init()
{
    threadWrite = NULL;
    threadRead  = NULL;

    cache = NULL;
}

//-------------------------------------------------------------------------------------------------
// Private functions
//-------------------------------------------------------------------------------------------------

void WControllerFilePrivate::createThreadWrite()
{
    if (threadWrite) return;

    Q_Q(WControllerFile);

    threadWrite = new WThreadActions(q);
}

void WControllerFilePrivate::createThreadRead()
{
    if (threadRead) return;

    Q_Q(WControllerFile);

    threadRead = new WThreadActions(q);
}

//-------------------------------------------------------------------------------------------------

void WControllerFilePrivate::registerLocalObject(WLocalObject * object)
{
    objects.append(object);
}

void WControllerFilePrivate::unregisterLocalObject(WLocalObject * object)
{
    objects.removeOne(object);
}

//-------------------------------------------------------------------------------------------------

void WControllerFilePrivate::registerFileWatcher(WFileWatcher * watcher)
{
    watchers.append(watcher);

    if (watchers.count() == 1)
    {
        Q_Q(WControllerFile);

        QObject::connect(&timer, SIGNAL(timeout()), q, SLOT(onCheckWatchers()));

        timer.start(1000);
    }
}

void WControllerFilePrivate::unregisterFileWatcher(WFileWatcher * watcher)
{
    watchers.removeOne(watcher);

    if (watchers.isEmpty())
    {
        Q_Q(WControllerFile);

        QObject::disconnect(&timer, 0, q, 0);

        timer.stop();
    }
}

//-------------------------------------------------------------------------------------------------

bool WControllerFilePrivate::isLoading() const
{
    foreach (WLocalObject * object, objects)
    {
        if (object->isProcessing()) return true;
    }

    return false;
}

//-------------------------------------------------------------------------------------------------
// Private static functions
//-------------------------------------------------------------------------------------------------

/* static */ bool WControllerFilePrivate::tryOpen(const QtLockedFile & file)
{
    QTimer timer;

    timer.start(CONTROLLERFILE_TIMEOUT);

    while (file.isLocked() && timer.isActive());

    if (file.isLocked())
    {
         return false;
    }
    else return true;
}

//-------------------------------------------------------------------------------------------------

/* static */ void WControllerFilePrivate::deleteDir(QDir & dir, bool recursive)
{
    QFileInfoList list;

    if (recursive)
    {
        list = dir.entryInfoList(QDir::NoDotAndDotDot | QDir::Files  | QDir::AllDirs
                                                      | QDir::System | QDir::Hidden);

        foreach (QFileInfo info, list)
        {
            if (info.isDir())
            {
                qDebug("Deleting folder %s", info.fileName().C_STR);

                WControllerFile::deleteFolder(info.absoluteFilePath(), true);
            }
            else
            {
                qDebug("Deleting file %s", info.fileName().C_STR);

                dir.remove(info.absoluteFilePath());
            }
        }
    }
    else
    {
        list = dir.entryInfoList(QDir::NoDotAndDotDot | QDir::Files
                                                      | QDir::System | QDir::Hidden);

        foreach (QFileInfo info, list)
        {
            qDebug("Deleting file %s", info.fileName().C_STR);

            dir.remove(info.absoluteFilePath());
        }
    }
}

//-------------------------------------------------------------------------------------------------
// Private slots
//-------------------------------------------------------------------------------------------------

void WControllerFilePrivate::onCheckWatchers()
{
    int fileCount = 20;

    foreach (WFileWatcher * watcher, watchers)
    {
        if (watcher->d_func()->checked) continue;

        if (watcher->d_func()->checkChange(fileCount) == false) return;
    }

    foreach (WFileWatcher * watcher, watchers)
    {
        watcher->d_func()->resetCheck();
    }
}

//=================================================================================================
// WControllerFile
//=================================================================================================
// Private

WControllerFile::WControllerFile() : WController(new WControllerFilePrivate(this)) {}

//-------------------------------------------------------------------------------------------------
// Initialize
//-------------------------------------------------------------------------------------------------

/* virtual */ void WControllerFile::init()
{
    Q_D(WControllerFile); d->init();
}

//-------------------------------------------------------------------------------------------------
// Interface
//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE */ WCacheFile * WControllerFile::getFile(const QString & url, QObject * parent,
                                                                             int       maxHost)
{
    Q_D(WControllerFile);

    if (d->cache)
    {
         return d->cache->getFile(url, parent, maxHost);
    }
    else return NULL;
}

/* Q_INVOKABLE */ WCacheFile * WControllerFile::getHttp(const QString & url, QObject * parent,
                                                                             int       maxHost)
{
    Q_D(WControllerFile);

    if (d->cache && WControllerNetwork::urlIsHttp(url))
    {
         return d->cache->getFile(url, parent, maxHost);
    }
    else return NULL;
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE */ QString WControllerFile::getFileUrl(const QString & url)
{
    Q_D(WControllerFile);

    if (d->cache)
    {
         return d->cache->getFileUrl(url);
    }
    else return QString();
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE */ WCacheFile * WControllerFile::writeCache(const QString    & url,
                                                           const QByteArray & array,
                                                           const QString    & extension,
                                                           QObject          * parent)
{
    Q_D(WControllerFile);

    if (d->cache)
    {
         return d->cache->writeFile(url, array, extension, parent);
    }
    else return NULL;
}

/* Q_INVOKABLE */ void WControllerFile::addCache(const QString & url, const QByteArray & array,
                                                                      const QString    & extension)
{
    Q_D(WControllerFile);

    if (d->cache)
    {
        d->cache->addFile(url, array, extension);
    }
    else qWarning("WControllerFile::addFile: Cannot add file %s. No cache.", url.C_STR);
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE */ void WControllerFile::waitActions()
{
    Q_D(WControllerFile);

    QTimer timer;

    timer.start(10000); // 10 seconds

    while (d->isLoading() && timer.isActive())
    {
        qApp->processEvents();
    }
}

//-------------------------------------------------------------------------------------------------
// Threaded write and read

WAbstractThreadReply * WControllerFile::startWriteAction(WAbstractThreadAction * action)
{
    Q_D(WControllerFile);

    d->createThreadWrite();

    return d->threadWrite->pushAction(action);
}

WAbstractThreadReply * WControllerFile::startReadAction(WAbstractThreadAction * action)
{
    Q_D(WControllerFile);

    d->createThreadRead();

    return d->threadRead->pushAction(action);
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startWriteFiles(const QStringList       & fileNames,
                                                        const QList<QByteArray> & datas)
{
    if (fileNames.isEmpty() || fileNames.count() != datas.count()) return NULL;

    WControllerFileWrite * action = new WControllerFileWrite;

    action->fileNames = fileNames;
    action->datas     = datas;

    startWriteAction(action);

    return action->controllerReply();
}

WControllerFileReply * WControllerFile::startRenameFiles(const QStringList & oldPaths,
                                                         const QStringList & newPaths)
{
    if (oldPaths.isEmpty() || oldPaths.count() != newPaths.count()) return NULL;

    WControllerFileRename * action = new WControllerFileRename;

    action->oldPaths = oldPaths;
    action->newPaths = newPaths;

    startWriteAction(action);

    return action->controllerReply();
}

WControllerFileReply * WControllerFile::startCopyFiles(const QStringList & fileNames,
                                                       const QStringList & newNames)
{
    if (fileNames.isEmpty() || fileNames.count() != newNames.count()) return NULL;

    WControllerFileCopy * action = new WControllerFileCopy;

    action->fileNames = fileNames;
    action->newNames  = newNames;

    startWriteAction(action);

    return action->controllerReply();
}

WControllerFileReply * WControllerFile::startDeleteFiles(const QStringList & paths)
{
    if (paths.isEmpty()) return NULL;

    WControllerFileDelete * action = new WControllerFileDelete;

    action->paths = paths;

    startWriteAction(action);

    return action->controllerReply();
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startCreateFolders(const QStringList & paths)
{
    if (paths.isEmpty()) return NULL;

    WControllerFileCreateFolders * action = new WControllerFileCreateFolders;

    action->paths = paths;

    startWriteAction(action);

    return action->controllerReply();
}

WControllerFileReply * WControllerFile::startDeleteFolders(const QStringList & paths,
                                                           bool                recursive)
{
    if (paths.isEmpty()) return NULL;

    WControllerFileDeleteFolders * action = new WControllerFileDeleteFolders;

    action->paths     = paths;
    action->recursive = recursive;

    startWriteAction(action);

    return action->controllerReply();
}

WControllerFileReply * WControllerFile::startDeleteFoldersContent(const QStringList & paths,
                                                                  bool                recursive)
{
    if (paths.isEmpty()) return NULL;

    WControllerFileDeleteFoldersContent * action = new WControllerFileDeleteFoldersContent;

    action->paths     = paths;
    action->recursive = recursive;

    startWriteAction(action);

    return action->controllerReply();
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startCreatePaths(const QStringList & paths)
{
    if (paths.isEmpty()) return NULL;

    WControllerFileCreatePaths * action = new WControllerFileCreatePaths;

    action->paths = paths;

    startWriteAction(action);

    return action->controllerReply();
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startWriteFile(const QString    & fileName,
                                                       const QByteArray & data)
{
    return startWriteFiles(QStringList() << fileName, QList<QByteArray>() << data);
}

WControllerFileReply * WControllerFile::startRenameFile(const QString & oldPath,
                                                        const QString & newPath)
{
    return startRenameFiles(QStringList() << oldPath, QStringList() << newPath);
}

WControllerFileReply * WControllerFile::startCopyFile(const QString & fileName,
                                                      const QString & newName)
{
    return startCopyFiles(QStringList() << fileName, QStringList() << newName);
}

WControllerFileReply * WControllerFile::startDeleteFile(const QString & path)
{
    return startDeleteFiles(QStringList() << path);
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startCreateFolder(const QString & path)
{
    return startCreateFolders(QStringList() << path);
}

WControllerFileReply * WControllerFile::startDeleteFolder(const QString & path,
                                                          bool            recursive)
{
    return startDeleteFolders(QStringList() << path, recursive);
}

WControllerFileReply * WControllerFile::startDeleteFolderContent(const QString & path,
                                                                 bool            recursive)
{
    return startDeleteFoldersContent(QStringList() << path, recursive);
}

//-------------------------------------------------------------------------------------------------

WControllerFileReply * WControllerFile::startCreatePath(const QString & path)
{
    return startCreatePaths(QStringList() << path);
}

//-------------------------------------------------------------------------------------------------
// Static functions
//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::absolute(const QUrl & url)
{
    return absolute(url.toString());
}

/* Q_INVOKABLE static */ QString WControllerFile::absolute(const QString & string)
{
    QFileInfo info(string);

    return info.absoluteFilePath();
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::fileUrl(const QUrl & url)
{
    return QUrl::fromLocalFile(url.toString()).toString();
}

/* Q_INVOKABLE static */ QString WControllerFile::fileUrl(const QString & string)
{
    return QUrl::fromLocalFile(string).toString();
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::filePath(const QUrl & url)
{
    QString string = url.toString();

    if (string.startsWith("file://"))
    {
         return url.toLocalFile();
    }
    else return string;
}

/* Q_INVOKABLE static */ QString WControllerFile::filePath(const QString & string)
{
    if (string.startsWith("file://"))
    {
         return QUrl(string).toLocalFile();
    }
    else return string;
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::fileAbsolute(const QUrl & url)
{
    return fileAbsolute(url.toString());
}

/* Q_INVOKABLE static */ QString WControllerFile::fileAbsolute(const QString & string)
{
    return fileUrl(absolute(string));
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::currentPath(const QString & fileName)
{
    return QDir::currentPath() + '/' + fileName;
}

/* Q_INVOKABLE static */ QString WControllerFile::currentFileUrl(const QString & fileName)
{
    return fileUrl(currentPath(fileName));
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::applicationPath(const QString & fileName)
{
    return QCoreApplication::applicationDirPath() + '/' + fileName;
}

/* Q_INVOKABLE static */ QString WControllerFile::applicationFileUrl(const QString & fileName)
{
    return fileUrl(applicationPath(fileName));
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::folderPath(const QString & fileName)
{
    QFileInfo info(filePath(fileName));

    if (info.isDir())
    {
         return info.absoluteFilePath();
    }
    else return info.absolutePath();
}

/* Q_INVOKABLE static */ QString WControllerFile::folderUrl(const QString & fileName)
{
    return fileUrl(folderPath(fileName));
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::folderName(const QString & fileName)
{
    int index = fileName.lastIndexOf('/');

    if (index == -1)
    {
         return fileName;
    }
    else return fileName.mid(index + 1);
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QString WControllerFile::toLocalFile(const QString & url)
{
    if (url.toLower().startsWith("qrc"))
    {
        return url.mid(3);
    }

    QString string = QUrl(url).toLocalFile();

    if (string.isEmpty())
    {
         return url;
    }
    else return string;
}

/* Q_INVOKABLE static */ QString WControllerFile::toString(const QString & url)
{
    if (url.toLower().startsWith("qrc"))
    {
         return url.mid(3);
    }
    else return url;
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ QByteArray WControllerFile::readAll(const QString & fileName)
{
    QFile file(toLocalFile(fileName));

    if (file.open(QIODevice::ReadOnly) == false)
    {
        qWarning("WControllerFile::readAll: Failed to open file %s.", fileName.C_STR);

        return QByteArray();
    }

    return file.readAll();
}

//-------------------------------------------------------------------------------------------------

/* Q_INVOKABLE static */ bool WControllerFile::exists(const QString & fileName)
{
    QFile file(toLocalFile(fileName));

    return file.exists();
}

/* Q_INVOKABLE static */ bool WControllerFile::tryAppend(const QString & fileName)
{
    QFile file(toLocalFile(fileName));

    if (file.exists())
    {
         return file.open(QIODevice::Append);
    }
    else return false;
}

//-------------------------------------------------------------------------------------------------
// Files

/* static */ bool WControllerFile::writeFile(const QString & fileName, const QByteArray & data)
{
    QtLockedFile file(fileName);

    if (WControllerFilePrivate::tryOpen(file) == false)
    {
        qWarning("WControllerFile::writeFile: File is locked %s.", fileName.C_STR);

        return false;
    }

    if (file.open(QIODevice::WriteOnly) == false)
    {
        qWarning("WControllerFile::writeFile: Cannot open file %s.", fileName.C_STR);

        return false;
    }

    file.write(data);

    return true;
}

/* static */ bool WControllerFile::renameFile(const QString & oldPath, const QString & newPath)
{
    QtLockedFile file(oldPath);

    if (WControllerFilePrivate::tryOpen(file) == false)
    {
        qWarning("WControllerFile::renameFile: File is locked %s.", oldPath.C_STR);

        return false;
    }

    return file.rename(newPath);
}

/* static */ bool WControllerFile::copyFile(const QString & fileName, const QString & newName)
{
    QtLockedFile file(newName);

    if (file.exists())
    {
        if (WControllerFilePrivate::tryOpen(file) == false)
        {
            qWarning("WControllerFile::copyFile: File is locked %s.", newName.C_STR);

            return false;
        }
    }

    return QFile::copy(fileName, newName);
}

/* static */ bool WControllerFile::deleteFile(const QString & fileName)
{
    QtLockedFile file(fileName);

    if (WControllerFilePrivate::tryOpen(file) == false)
    {
        qWarning("WControllerFile::deleteFile: File is locked %s.", fileName.C_STR);

        return false;
    }

    return file.remove();
}

//-------------------------------------------------------------------------------------------------
// Directories

/* static */ bool WControllerFile::createFolder(const QString & path)
{
    return QDir().mkdir(path);
}

/* static */ bool WControllerFile::moveFolder(const QString & oldPath, const QString & newPath)
{
    return QDir().rename(oldPath, newPath);
}

/* static */ bool WControllerFile::deleteFolder(const QString & path, bool recursive)
{
     QDir dir(path);

     if (dir.exists() == false) return false;

     WControllerFilePrivate::deleteDir(dir, recursive);

     return dir.rmdir(dir.absolutePath());
}

/* static */ bool WControllerFile::deleteFolderContent(const QString & path, bool recursive)
{
    QDir dir(path);

    if (dir.exists() == false) return false;

    WControllerFilePrivate::deleteDir(dir, recursive);

    return true;
}

//-------------------------------------------------------------------------------------------------

/* static */ QFileInfoList WControllerFile::recursiveEntryInfoList(const QString & path)
{
     QDir dir(path);

     QFileInfoList list = dir.entryInfoList(QDir::Dirs | QDir::NoDotAndDotDot);

     foreach (QFileInfo info, list)
     {
        list.append(recursiveEntryInfoList(info.filePath()));
     }

     return list;
}

//-------------------------------------------------------------------------------------------------
// Private slots
//-------------------------------------------------------------------------------------------------

void WControllerFile::onCheckWatchers()
{
    Q_D(WControllerFile);

    int fileCount = 20;

    foreach (WFileWatcher * watcher, d->watchers)
    {
        if (watcher->d_func()->checked) continue;

        if (watcher->d_func()->checkChange(fileCount) == false) return;
    }

    foreach (WFileWatcher * watcher, d->watchers)
    {
        watcher->d_func()->resetCheck();
    }
}

//-------------------------------------------------------------------------------------------------
// Properties
//-------------------------------------------------------------------------------------------------

QString WControllerFile::pathStorage() const
{
    Q_D(const WControllerFile); return d->pathStorage;
}

void WControllerFile::setPathStorage(const QString & path)
{
    Q_D(WControllerFile);

    if (d->pathStorage == path) return;

    d->pathStorage = path;

    emit pathStorageChanged();
}

//-------------------------------------------------------------------------------------------------

WCache * WControllerFile::cache() const
{
    Q_D(const WControllerFile); return d->cache;
}

void WControllerFile::setCache(WCache * cache)
{
    Q_D(WControllerFile);

    if (d->cache == cache) return;

    if (d->cache) delete d->cache;

    d->cache = cache;

    if (cache) cache->setParent(this);

    emit cacheChanged();
}

#endif // SK_NO_CONTROLLERFILE

#include "WControllerFile.moc"
